#!/usr/bin/env python
# Purpose:
#      Scan fortran sources, looking for use, include and
#      module statements. 
#      Output dependencies for use in Make.
#      On request, also output macro definitions
#      try:
#        ./makedepo.py
#
import sys
import re
import getopt
import os


def sortu(a):
    if len(a) == 0:
        return a
    x = a.split()
    x.sort()
    l = [x[0]]
    u = x[0]
    for p in x[1:]:
        if l[-1] != p:
            l = l + [p]
            u = u + " " + p
    return u


def removenil(x):
    u = {}
    for i in x:
        if x[i]:
            u[i] = x[i]
    return u


def deps(mode):
    if mode == "M:F":          # m1.mod m2.mod: sub.f90
        provn = removenil(prov1)
        for f in provn:
            print(provn[f] + ": " + f)
        return

    if mode == "M:O":          # m1.mod m2.mod: sub.o
        provn = removenil(prov1)
        for f in provn:
            print(provn[f] + ": " + crs[f])
        return

    for f in src:
        if mode == "O:FHO":    # sub.o: sub.f90 inc1.h inc2.h sub1.o sub2.o
            rr = req[f].split()
            d = ""
            for r in rr:
                try:
                    if prov[r] != f:
                        d = d + " " + prov[r]
                except KeyError:
                    pass
            print(f + ": " + src[f] + " " + inc[f] + " " + sortu(d))
        elif mode == "O:FH":         # sub.o: sub.f90 inc1.h inc2.h
            print(f + ": " + src[f] + " " + inc[f])
        elif mode == "O:M":          # sub.o: m1.mod m2.mod
            # use only modulefiles that are generated in the sources
            # skip modulefiles that are generated in the source itself
            d = ""
            if(req[f]):
                rr = req[f].split()
                for r in rr:
                    try:
                        if prov[r] != f:
                            d = d + " " + r
                    except KeyError:
                        pass
            if d:
                print(f + ": " + sortu(d))
        elif mode == "O:FHM":         # sub.o: sub.f90 inc.h m1.mod m2.mod
            d = src[f] + " " + inc[f]
            if(req[f]):
                rr = req[f].split()
                for r in rr:
                    try:
                        if prov[r] != f:
                            d = d + " " + r
                    except KeyError:
                        pass
            print(f + ": " + sortu(d))

# start of main program

prov = {}   # providers {'sub1.mod':'sub.o', ...}
prov1 = {}  # providers {'sub.f90':'sub1.mod sub2.mod ...', ...}
req = {}    # requesters   {'sub.o':'a1.mod a2.mod ...',...}
inc = {}    # requesters   {'sub.o':'inc1.h inc2.h ...',...}
src = {}    # object:source {'sub.o':'sub.f90', ...}
crs = {}    # source:object {'sub.f90':'sub.o', ...}

findsuffix = re.compile(r'.[^.]*$')
findquote = re.compile(r'[\'"]')

modesallowed = ["O:FHO", "O:M", "O:FH", "M:F","M:O", "O:FHM"]

def usage():
    print >> sys.stderr,"""
Usage:
    makedepo.py [-m mode] [-p prefix] [-s suffix] [-y] fortransources
     mode           ouput

      M:F      m1.mod m2.mod: sub.f90
      M:O      m1.mod m2.mod: sub.suffix 
      O:FHO    sub.o: sub.f90 inc.h sub1.suffix sub2.suffix ..
      O:FH     sub.o: sub.f90 inc1.h inc2.h ...
      O:M      sub.o: m1.mod m2.mod ...
      O:FHM    sub.o: sub.f90 inc1.h inc2.h ... m1.mod m2.mod ...

      default: O:FHO
  Note: O:FHO is computed based on modules

   prefix: this will be prefixed to sub1.o, default ""
   suffix: use this suffix for object files, default ".o"

   """

try:
    optlist, args = getopt.getopt(sys.argv[1:], 'm:s:p:y')
except:
    usage()
    sys.exit(1)

if (len(args) == 0):
    usage()
    sys.exit(1)

modes = ["O:FHO"]
prefix = ""
suffix = ".o"
genmacro = False


for o, a in optlist:
    if o == "-m":
        modes = a.split(",")
    elif o == "-p":
        prefix = a
    elif o == "-s":
        suffix = a

for mode in modes:
    if not mode in modesallowed:
        print "Mode ", mode, " not allowed"
        sys.exit(1)

for f in args:
    try:
        srcfile = open(f)
        basename = re.sub(findsuffix, '', f)
        oname = prefix + basename + suffix
        src[oname] = f
        crs[f] = oname
        req[oname] = ""
        prov1[f] = ""
        inc[oname] = ""
        for line in srcfile:
            line = line.strip().split()
            word1 = ""
            if len(line):
                word1 = line[0]
            word2 = ""
            if len(line) > 1:
                word2 = line[1]
            if word1.lower() == "module":
                if word2.lower() == "procedure":
                    continue
                prov[word2.lower() + ".mod"] = oname
                if prov1[f]:
                    prov1[f] = prov1[f] + " " + word2.lower() + ".mod"
                else:
                    prov1[f] = word2.lower() + ".mod"
            if word1.lower() == "use":
                cp = word2.rfind(",")
                if cp != -1:
                    word2 = word2[:cp]
                req[oname] = req[oname] + " " + word2.lower() + ".mod"
            if word1 == "include":
                # do not use dependencies of files that are not visible
                word2 = re.sub(findquote, '', word2)
                if os.path.exists(word2):
                # if 1:
                    inc[oname] = inc[oname] + " " + word2
        srcfile.close()
    except IOError:
        pass


# inc  = removenil(inc)

for i in inc:
    inc[i] = sortu(inc[i])

for i in prov:
    prov[i] = sortu(prov[i])

for i in req:
    req[i] = sortu(req[i])

if 0:
    print "src",src
    print "prov",prov
    print "prov1",prov1
    print "req",req
    print "inc",inc
for mode in modes:
    deps(mode)

if genmacro:
    printmacros()
